# Copyright (C) 2022-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for :mod:`cieclib.sessions`."""

import logging

import pytest

from .. import sessions as ciecplib_sessions

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class TestSession():
    TEST_CLASS = ciecplib_sessions.Session

    def test_init_error(self):
        """Test `init` error handling."""
        with pytest.raises(ValueError) as exc:
            self.TEST_CLASS(idp=None, kerberos=False)
        assert str(exc.value).startswith("no Identity Provider")

    @pytest.mark.parametrize("debug", (False, True))
    def test_debug(self, debug):
        """Check that the ``debug`` keyword argument is handled properly."""
        sess = self.TEST_CLASS(
            idp="https://example.com/idp/profile/SAML2/SOAP/ECP",
            kerberos=False,
            debug=debug,
        )
        # check that the parameter was recorded properly
        assert sess.debug is debug
        # check that the level was propagated up to the root logger properly
        assert logging.getLogger().isEnabledFor(logging.DEBUG) is debug
        # check that closing does the right thing
        sess.close()
        assert (
            # either debug is disabled
            logging.getLogger().isEnabledFor(logging.DEBUG) is False
            # or the logger was set back to NOTSET
            or logging.getLogger().getEffectiveLevel() == logging.NOTSET
        )



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`ciecplib.tool.ecp_get_cert`."""

from unittest import mock

import pytest

from ... import x509 as ciecplib_x509
from .. import ecp_get_cert


@mock.patch("os.unlink", mock.Mock(side_effect=(None, FileNotFoundError)))
@pytest.mark.parametrize("dummy", (True, False))
def test_destroy(dummy):
    """Check that the --destroy option works whether the file exists or not."""
    ecp_get_cert.main(["--destroy"])


def test_ecp_get_cert(tmp_path, x509, private_key):
    """Check that a standard workflow correctly writes the key it received."""
    x509_path = tmp_path / "x509.pem"
    with mock.patch(
        "ciecplib.tool.ecp_get_cert.get_cert",
        return_value=(x509, private_key),
    ):
        ecp_get_cert.main([
            "--file", str(x509_path),
            "--identity-provider", "test",
        ])
    assert ciecplib_x509.load_cert(x509_path) == x509


@mock.patch("ciecplib.tool.ecp_get_cert.get_cert")
def test_ecp_get_cert_reuse(get_cert, x509_path):
    """Check that the --reuse option works if the key validates."""
    # reuse an existing certificate file
    ecp_get_cert.main([
        "--file", str(x509_path),
        "--identity-provider", "test",
        "--reuse",
    ])
    # make sure that it did get reused
    get_cert.assert_not_called()


def test_ecp_get_cert_reuse_fail(x509_path, x509, private_key):
    """Check that the --reuse option works if the key fails to validate."""
    bad = x509_path.with_suffix(".bad")
    with mock.patch(
        "ciecplib.tool.ecp_get_cert.get_cert",
        return_value=(x509, private_key),
    ) as mock_get_cert:
        # attempt to reuse an certificate file that doesn't exist
        ecp_get_cert.main([
            "--file", str(bad),
            "--identity-provider", "test",
            "--reuse",
        ])
        # make sure that it did get reused
        mock_get_cert.assert_called_once()
    assert ciecplib_x509.load_cert(bad) == x509



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

import pytest

from .. import env as ciecplib_env

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize("env, result", [
    ({}, None),
    ({"ECP_IDP": "TEST"}, "TEST"),
    ({"CIGETCERTOPTS": "-i TEST2"}, "TEST2"),
])
@mock.patch.dict("os.environ", clear=True)
def test_get_default_idp(env, result):
    os.environ.update(env)
    assert ciecplib_env._get_default_idp() == result



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""HTTP request method with end-to-end ECP authentication."""

import sys
from functools import wraps
from textwrap import indent
try:
    from contextlib import nullcontext
except ImportError:  # python < 3.7
    class nullcontext:
        def __init__(self, enter_result=None):
            self.enter_result = enter_result

        def __enter__(self):
            return self.enter_result

        def __exit__(self, *excinfo):
            pass

from .env import DEFAULT_IDP
from .sessions import Session

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _ecp_session(func):
    """Decorate a function to open a `requests_ecp.Session`.

    If a `Session` is already open, this function returns an instance of
    `contextlib.nullcontext` that should _only_ be used with the `with`
    statement, not a real `Session` object. This is only to support chaining
    this method reentrant with respect to setting/resetting debug logging
    levels.
    """
    @wraps(func)
    def _wrapper(*args, **kwargs):
        sess = kwargs.pop("session", None)
        if sess is None:
            sess = Session(
                idp=kwargs.pop("endpoint", DEFAULT_IDP),
                username=kwargs.pop("username", None),
                password=kwargs.pop("password", None),
                kerberos=kwargs.pop("kerberos", None),
                cookiejar=kwargs.pop("cookiejar", None),
                debug=kwargs.get("debug", False),
            )
        else:
            sess = nullcontext(enter_result=sess)
        return func(*args, session=sess, **kwargs)
    return _wrapper


def _session_func_factory(method, docstring):
    @_ecp_session
    def _func(url, **kwargs):
        kwargs.pop("debug", None)  # not supported by requests functions
        # the decorator guarantees us a session
        with kwargs.pop("session") as session:
            meth = getattr(session, method)
            return meth(url, **kwargs)

    # post-process the docstring
    doc = f"""{docstring}

{_request_params_doc.format(method=method)}
"""
    if sys.version_info < (3, 13, 0):
        # older versions don't have https://github.com/python/cpython/issues/81283
        doc = indent(doc, "    ").strip()
    _func.__doc__ = doc

    return _func


_request_params_doc = """
Parameters
----------
url : `str`
    URL path for request.

endpoint : `str`, optional
    ECP endpoint name or URL for request.

username : `str`, optional
    the username with which to authenticate, will be prompted for
    if not given, and not using ``kerberos``.

password : `str`, optional
    the password with which to authenticate, will be prompted for
    if not given, and not using ``kerberos``.

kerberos : `bool`, optional
    use existing kerberos credential for login, default is to try, but
    fall back to username/password prompt.

cookiejar : `http.cookiejar.CookieJar`, optional
    a jar of cookies to add to the `requests.Session`.

kwargs
    other keyword arguments are passed directly to
    :meth:`requests.Session.{method}`

Returns
-------
response : `http.client.HTTPResponse`
    the response from the URL

See Also
--------
requests.Session.{method}
    for full details of the request handling
""".strip()

get = _session_func_factory(
    "get",
    "Send a GET request using ECP authentication."
)

head = _session_func_factory(
    "head",
    "Send a HEAD request using ECP authentication."
)

post = _session_func_factory(
    "post",
    "Send a POST request using ECP authentication."
)



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for :mod:`cieclib.x509`."""

import sys
from io import StringIO
from unittest import mock

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

import pytest

from .. import x509 as ciecplib_x509

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

X509_SUBJECT = (
    "/CN=albert einstein/C=UK/ST=Wales/L=Cardiff/"
    "O=Cardiff University/OU=Gravity"
)


def test_load_cert(x509, x509_path):
    cert = ciecplib_x509.load_cert(x509_path)
    assert cert == x509


def test_load_pkcs12(x509, private_key):
    """Test that `ciecplib.x509.load_pkcs12` returns equivalent certs and keys."""
    from cryptography.hazmat.primitives.serialization import (
        BestAvailableEncryption,
    )
    try:
        from cryptography.hazmat.primitives.serialization.pkcs12 import (
            serialize_key_and_certificates,
        )
    except ImportError:  # cryptography < 3.0
        from OpenSSL import crypto
        p12obj = crypto.PKCS12()
        p12obj.set_certificate(crypto.X509.from_cryptography(x509))
        p12obj.set_privatekey(crypto.PKey.from_cryptography_key(private_key))
        p12 = p12obj.export(b"password")
    else:
        p12 = serialize_key_and_certificates(
            b"test",
            private_key,
            x509,
            None,
            BestAvailableEncryption(b"password"),
        )
    cert, key = ciecplib_x509.load_pkcs12(p12, b"password")
    assert cert == x509
    assert key.private_numbers() == private_key.private_numbers()


def test_time_left(x509):
    assert ciecplib_x509.time_left(x509) <= 86400


@mock.patch("ciecplib.x509.time_left", mock.MagicMock(return_value=4000))
def test_check_cert(x509):
    ciecplib_x509.check_cert(x509, hours=1, proxy=False, rfc3820=False)


@mock.patch("ciecplib.x509.time_left", mock.MagicMock(return_value=4000))
def test_check_cert_time(x509):
    with pytest.raises(RuntimeError):
        ciecplib_x509.check_cert(x509, hours=2)


@pytest.mark.parametrize("proxy, certtype", [
    (True, "end entity credential"),
    (False, "some sort of proxy"),
])
@mock.patch("ciecplib.x509._cert_type")
@mock.patch("ciecplib.x509.time_left", mock.MagicMock(return_value=4000))
def test_check_cert_proxy_errors(mock_cert_type, x509, proxy, certtype):
    mock_cert_type.return_value = certtype
    with pytest.raises(RuntimeError):
        ciecplib_x509.check_cert(x509, hours=1, proxy=proxy)


@mock.patch(
    "ciecplib.x509._cert_type",
    mock.MagicMock(return_value="legacy globus proxy"),
)
@mock.patch("ciecplib.x509.time_left", mock.MagicMock(return_value=4000))
def test_check_cert_rfc3820_error(x509):
    with pytest.raises(RuntimeError):
        ciecplib_x509.check_cert(x509, hours=1, rfc3820=True)


@pytest.mark.parametrize("timeleft", (
    pytest.param(0, id="expired"),
    pytest.param(10, id="active"),
))
@mock.patch("ciecplib.x509.time_left")
def test_print_cert_info(mock_time_left, timeleft, x509, capsys):
    mock_time_left.return_value = timeleft
    ciecplib_x509.print_cert_info(
        x509,
        path="test",
        verbose=True,
        stream=sys.stdout,
    )
    out = capsys.readouterr().out.strip()
    assert out.startswith(f"subject  : {X509_SUBJECT}")
    assert "path     : test" in out
    assert ("[EXPIRED]" in out) is (timeleft == 0)
    assert out.endswith("-----END CERTIFICATE-----")


@pytest.mark.parametrize(("display", "result"), [
    (("subject",), X509_SUBJECT),
    (("path", "issuer"), f"test\n{X509_SUBJECT}"),
    (("issuer", "path"), f"{X509_SUBJECT}\ntest"),
])
def test_print_cert_info_display(x509, display, result):
    stream = StringIO()
    ciecplib_x509.print_cert_info(
        x509,
        path="test",
        display=display,
        stream=stream,
    )
    stream.seek(0)
    out = stream.read()
    assert out.strip() == result.strip()


@pytest.mark.parametrize("proxy", (False, True))
def test_write_cert(tmp_path, private_key, x509, proxy):
    path = tmp_path / "509.pem"
    # write the p12 cert to a file
    ciecplib_x509.write_cert(path, x509, private_key, use_proxy=proxy)
    # check that we can read it
    new = ciecplib_x509.load_cert(path)
    # check that we get the same cert back
    if proxy:
        assert new.issuer == x509.subject
        assert ciecplib_x509._x509_name_str(new.subject).startswith(
            ciecplib_x509._x509_name_str(x509.subject),
        )
        assert ciecplib_x509._not_valid_after(new) \
               <= ciecplib_x509._not_valid_after(x509)
        assert ciecplib_x509._not_valid_before(new) \
               >= ciecplib_x509._not_valid_before(x509)
        assert new.extensions
    else:
        assert new == x509


@pytest.mark.parametrize("limited, ctype", [
    (False, "RFC 3820 compliant impersonation proxy"),
    (True, "RFC 3820 compliant limited proxy"),
])
def test_generate_proxy(x509, private_key, limited, ctype):
    """Test that :func:`ciecplib.x509.generate_proxy` generates a good proxy."""
    # generate a proxy certificate
    proxy, pkey = ciecplib_x509.generate_proxy(
        x509,
        private_key,
        limited=limited,
    )

    # check that the (correct) private_key was used to sign the proxy
    private_key.public_key().verify(
        proxy.signature,
        proxy.tbs_certificate_bytes,
        padding.PKCS1v15(),
        algorithm=hashes.SHA256(),
    )

    # check that the cerificate type matches
    assert ciecplib_x509._cert_type(proxy) == ctype



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Create an X.509 certificate using ECP authentication.

ecp-get-cert enables you to authenticate using SAML/ECP against any
identity provider registered with CILogon, and create an X.509
credential for use in querying services under the same identity and
access management domain.
"""

from ..ui import get_cert
from ..x509 import (
    check_cert,
    load_cert,
    print_cert_info,
    write_cert,
)
from ..utils import DEFAULT_X509_USER_FILE
from .utils import (
    ArgumentParser,
    destroy_file,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

EPILOG = r"""
Examples:

    $ ecp-get-cert -i 'My Institution'

to authenticate with a username and password prompt, or

    $ ecp-get-cert -u user.name -i 'My Institution'

to authenticate with only a password prompt, or

    $ ecp-get-cert -i 'My Institution' -k

to reuse an existing kerberos (``kinit``) credential.

The identitity provider name can be given in a number of ways, so long as the
argument uniquely identifies a provider.  For example, the following are all
equivalent:

    $ ecp-get-cert -i 'Cardiff University'
    $ ecp-get-cert -i Cardiff
    $ ecp-get-cert -i idp.cf.ac.uk
    $ ECP_IDP="Cardiff" ecp-get-cert

Environment:

X509_USER_PROXY:
    the default path for the credential file
ECP_IDP:
    the name/url of the default Identity Provider (--institution)
"""

MANPAGE = [
    {"heading": "environment",
     "content": """
.TP
.B "X509_USER_PROXY"
The default path for the credential file """,
     },
]


def create_parser():
    """Create a command-line argument parser.

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        epilog=EPILOG,
        prog="ecp-get-cert",
        manpage=MANPAGE,
        add_auth=True,
        add_helpers=True,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="write debug output (uses both stderr and stdout, "
             "implies --verbose)",
    )
    parser.add_argument(
        "-f",
        "--file",
        default=DEFAULT_X509_USER_FILE,
        help="certificate file to create/reuse/destroy",
    )
    parser.add_argument(
        "-H",
        "--hours",
        type=int,
        default=277,
        help="lifetime of the certificate"
    )
    parser.add_argument(
        "-p",
        "--proxy",
        action="store_true",
        default=False,
        help="create RFC 3820 compliant impersonation proxy"
    )
    parser.add_argument(
        "-r",
        "--reuse",
        default=False,
        type=float,
        const=1.,
        nargs="?",
        metavar="HOURS",
        help="reuse an existing certificate if valid for more than "
             "%(const)s hours, or pass a number of hours to specify",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="write verbose output to stdout",
    )
    parser.add_argument(
        "-X",
        "--destroy",
        action="store_true",
        default=False,
        help="destroy existing certificate"
    )
    return parser


def parse_args(parser, args=None):
    """Parse and validate the command-line arguments.

    Returns
    -------
    args : `argparse.Namespace`
    """
    args = parser.parse_args(args=args)

    if args.debug:
        args.verbose = True

    return args


def can_reuse(path, proxy=None):
    try:
        check_cert(load_cert(path), proxy=proxy)
    except (RuntimeError, OSError):
        return False
    return True


def main(args=None):
    parser = create_parser()
    args = parse_args(parser, args=args)

    def vprint(*pargs, **kwargs):
        """Execute `print` only if --verbose was given."""
        if args.verbose:
            print(*pargs, **kwargs)

    # if asked to destroy, just do that
    if args.destroy:
        destroy_file(args.file, "credential file", verbose=args.verbose)
        return 0

    # if asked to reuse, check that we can
    if args.reuse:
        vprint("Validating existing certificate...", end=" ")
        args.reuse = can_reuse(args.file, proxy=args.proxy)
        if args.reuse:
            vprint("OK")
        else:
            vprint("failed, will regenerate")

    # get new certificate
    if not args.reuse:
        vprint("Fetching certificate...")
        cert, key = get_cert(
            endpoint=args.identity_provider,
            username=getattr(args, "username", None),
            kerberos=args.kerberos,
            hours=args.hours,
            debug=args.debug,
        )

        # write certificate to a file
        vprint("Storing certificate...")
        write_cert(
            args.file,
            cert,
            key,
            use_proxy=args.proxy,
            minhours=args.hours,
        )
        vprint("X.509 credential stored")

    # load the cert from file to print information
    if args.verbose:
        print_cert_info(
            load_cert(args.file),
            path=args.file,
            verbose=True,
        )


if __name__ == "__main__":
    main()



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""X.509 functions for CIECPLib."""

import calendar
import datetime
import shutil
import sys
import tempfile
import time

from cryptography import x509 as crypto_x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

PROXY_CERT_INFO_EXT_OID = crypto_x509.ObjectIdentifier("1.3.6.1.5.5.7.1.14")

# backport short names for cryptography < 2.5
_NAMEOID_TO_NAME = {
    crypto_x509.NameOID.COMMON_NAME: "CN",
    crypto_x509.NameOID.COUNTRY_NAME: "C",
    crypto_x509.NameOID.DOMAIN_COMPONENT: "DC",
    crypto_x509.NameOID.LOCALITY_NAME: "L",
    crypto_x509.NameOID.ORGANIZATION_NAME: "O",
    crypto_x509.NameOID.ORGANIZATIONAL_UNIT_NAME: "OU",
    crypto_x509.NameOID.STATE_OR_PROVINCE_NAME: "ST",
    crypto_x509.NameOID.STREET_ADDRESS: "STREET",
    crypto_x509.NameOID.USER_ID: "UID",
}


def _parse_unix_time(asn1time):
    """Parse an ASN1 unix time into a python `time` tuple."""
    return calendar.timegm(asn1time.get_datetime().timetuple())


def load_cert(path):
    """Load an X.509 certificate from file containing PEM-encoded data.

    Parameters
    ----------
    path : `str`, `pathlib.Path`
        the file path from which to read

    Returns
    -------
    cert : `cryptography.x509.Certificate`
        the parsed certificate
    """
    with open(str(path), "rb") as fobj:
        return crypto_x509.load_pem_x509_certificate(
            fobj.read(),
            backend=default_backend(),
        )


def load_pkcs12(raw, password):
    """Load an X.509 certificate and key from a PKCS12 blob."""
    try:
        from cryptography.hazmat.primitives.serialization.pkcs12 import (
            load_pkcs12 as _load_pkcs12,
        )
    except ImportError:  # cryptography < 36.0
        from OpenSSL.crypto import load_pkcs12 as _load_pkcs12
        p12 = _load_pkcs12(raw, password)
        return (
            p12.get_certificate().to_cryptography(),
            p12.get_privatekey().to_cryptography_key(),
        )

    p12 = _load_pkcs12(raw, password)
    return (
        p12.cert.certificate,
        p12.key,
    )


def _not_valid_before(cert, naive=False):
    """Return the `not_valid_before_utc` attribute of the given cert.

    Falling back to a timezone-aware replacement of the `not_valid_before`
    attribute for cryptography < 42.0.0.
    """
    try:
        return cert.not_valid_before_utc
    except AttributeError:  # cryptography < 42.0.0
        if naive:
            return cert.not_valid_before
        return cert.not_valid_before.replace(tzinfo=datetime.timezone.utc)


def _not_valid_after(cert, naive=False):
    """Return the `not_valid_after_utc` attribute of the given cert.

    Falling back to a timezone-aware replacement of the `not_valid_after`
    attribute for cryptography < 42.0.0.
    """
    try:
        return cert.not_valid_after_utc
    except AttributeError:  # cryptography < 42.0.0
        if naive:
            return cert.not_valid_after
        return cert.not_valid_after.replace(tzinfo=datetime.timezone.utc)


def time_left(cert):
    """Return the number of seconds left on this certificate.

    If the certificate has expired, ``0`` is returned.

    Parameters
    ----------
    cert : `cryptography.x509.Certificate`
        The certificate to inspect.
    """
    expiry = _not_valid_after(cert).timetuple()
    return max(0, int(calendar.timegm(expiry) - time.time()))


def _x509_name_str(obj):
    """Return the name of the x509 object as a string."""
    try:
        parts = [x.rfc4514_string() for x in obj.rdns]
    except AttributeError:  # cryptography < 2.5
        parts = [
            f"{_NAMEOID_TO_NAME[attr.oid]}={attr.value}"
            for rdn in obj.rdns
            for attr in rdn
        ]
    return f"/{'/'.join(parts)}"


def check_cert(cert, hours=1, proxy=None, rfc3820=True):
    """Validate an X509 certificate.

    Parameters
    ----------
    cert : `cryptography.x509.Certificate`
        the certificate object to check

    hours : `float`, optional
        minimum number of hours remaining before expiry

    proxy : `bool`, `None`, optional
        if `True` (`False`), validate that the certificate is (is not)
        an impersonation proxy, if `None` (default) don't check.

    rfc3820 : `bool`, optional
        if `True` assert that, if the certificate is a proxy, that it
        is RFC 3820 compliant
    """
    # check expiry
    remaining = time_left(cert)
    if remaining < hours * 3600.:
        raise RuntimeError(
            "less than {0} hours remaining on X509 certificate".format(hours)
        )

    # check proxy type
    ctype = _cert_type(cert)
    isproxy = "proxy" in ctype
    if proxy is True and not isproxy:
        raise RuntimeError("certificate is not a proxy")
    if proxy is False and isproxy:
        raise RuntimeError("certificate is a proxy")
    if rfc3820 and "legacy globus proxy" in ctype:
        raise RuntimeError("proxy certificate is not RFC 3820 compliant")


def _cert_type(x509):
    """Return the type of the given x509 certificate object."""
    # parse name entry as common name
    subject = _x509_name_str(x509.subject)
    ntype, name = subject.rsplit("/", 1)[-1].split("=", 1)

    # if name entry is not 'common name' then EEC
    if ntype != "CN":
        return "end entity credential"

    if name == "proxy":
        return "full legacy globus proxy"
    if name == "limited proxy":
        return "limited legacy globus proxy"

    # get policy language
    try:
        policy = _get_cert_policy_language(x509)
    except crypto_x509.ExtensionNotFound:
        return "end entity credential"
    except ValueError:  # no policy language
        return "unidentified proxy"
    else:
        if policy == "1.3.6.1.4.1.3536.1.1.1.9":
            return "RFC 3820 compliant limited proxy"
        if policy == "Inherit all":
            return "RFC 3820 compliant impersonation proxy"
        return "RFC 3820 compliant restricted proxy"


def _get_cert_policy_language(x509):
    """Parse the policy language from the proxyCertInfo extension.

    This is a terrible hack because pyca/cryptography doesn't actually support
    the proxyCertInfo extension.
    """
    ext = x509.extensions.get_extension_for_oid(PROXY_CERT_INFO_EXT_OID)
    for line in ext.value.value.split():
        if line.endswith(b"+\x06\x01\x05\x05\x07\x15\x01"):
            return "Inherit all"
        if line.endswith(b"+\x06\x01\x04\x01\x9bP\x01\x01\x01"):
            return "1.3.6.1.4.1.3536.1.1.1.9"
    raise ValueError("no policy language found in cert")


def print_cert_info(
    x509,
    path=None,
    display=None,
    verbose=True,
    stream=sys.stdout,
):
    """Print info about an X.509 certificate.

    Parameters
    ----------
    x509 : `cryptography.x509.Certificate`
        the certificate to parse

    path : `str`, optional
        the path of the certificate file on disk

    display : `list`, optional
        list of certificate parameters to display; if given each is displayed
        in order in plaintext

    verbose : `bool`, optional
        if `True` (default) print the full text of the certificate

    stream : `file`, optional
        the file object to print to, defaults to `sys.stdout`
    """
    if display is None:
        display = []
    # parse parameters of certificate
    pkey = x509.public_key()
    remaining = time_left(x509)
    if "timeleft" in display:  # if plaintext, print seconds
        timeleft = str(max(-1, int(remaining)))
    elif remaining:  # otherwise format nicely
        timeleft = str(datetime.timedelta(seconds=remaining))
    else:
        timeleft = "0:00:00 [EXPIRED]"
    params = {
        "subject": _x509_name_str(x509.subject),
        "issuer": _x509_name_str(x509.issuer),
        "type": _cert_type(x509),
        "strength": f"{pkey.key_size} bits",
        "path": path,
        "timeleft": timeleft,
    }
    if verbose or "text" in display:
        params["text"] = "\n" + x509.public_bytes(
            encoding=Encoding.PEM,
        ).decode("utf-8").strip()

    # use chose specific attributes, so just print them in plain text
    if display:
        for attr in display:
            print(params[attr].strip(), file=stream)
    else:
        for attr in params:
            print(f"{attr:9s}: {params[attr]}", file=stream)


def write_cert(path, cert, key, use_proxy=False, minhours=168):
    """Write a PKCS12 certificate archive to file in X509 format.

    Parameters
    ----------
    path : `str`, `pathlib.Path`
        the desired location of the final X509 file

    pkcs12 : `OpenSSL.crypto.PKCS12`
        a PKCS12 archive

    use_proxy : `bool`, optional
        if `True`, generate an impersonation proxy, otherwise generate
        a standard end entity credential certificate

    minhours : `int`, optional
        the minimum duration of the proxy certificate, only used if
        `proxy=True` is given
    """
    if use_proxy:
        # we will store the original certificate as part of the signing chain
        chain = [cert.public_bytes(encoding=Encoding.PEM)]
        # generate a self-signed proxy
        cert, key = generate_proxy(cert, key, minhours=minhours)
    else:
        chain = []

    # write the cert, key pair, and the signing chain (if any)
    blocks = [
        # the cert
        cert.public_bytes(encoding=Encoding.PEM),
        # the private key associated with the cert's public key
        key.private_bytes(
            encoding=Encoding.PEM,
            format=PrivateFormat.PKCS8,
            encryption_algorithm=NoEncryption(),
        ),
    ] + chain

    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        # write cert and key to temporary location
        for block in blocks:
            tmp.write(block)
        tmp.close()

        # move tmpfile into place
        shutil.move(tmp.name, str(path))


def generate_proxy(cert, key, minhours=168, limited=False, bits=2048):
    """Generate a proxy certificate based on a certificate.

    Parameters
    ----------
    cert : `cryptography.X509.Certificate`
        The certificate object.

    key : `cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey`
        The RSA key object used to sign the original certificate.

    minhours : `float`
        The minimum lifetime of the proxy certificate. This is bounded by
        the lifetime of the original certificate.

    limited : `bool`
        If `True`, generate a limited proxy.

    bits : `int`
        The number of bits (size) to use for the private key used to sign
        the proxy certificate.

    Returns
    -------
    proxycert : `cryptography.X509.Certificate`
        The proxy certificate.

    proxykey : `cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey`
        The RSA private key used to sign the proxy certificate.
    """
    # generate a new key pair with which to sign the proxy
    proxy_private_key = generate_private_key(
        public_exponent=65537,
        key_size=bits,
        backend=default_backend(),
    )
    proxy_public_key = proxy_private_key.public_key()

    # create a serial number for the proxy
    serial = crypto_x509.random_serial_number()

    # generate a new subject by appending the serial number to
    # the subject of the original certificate
    proxy_subject = crypto_x509.Name(
        list(cert.subject) + [
            crypto_x509.NameAttribute(
                crypto_x509.NameOID.COMMON_NAME,
                str(serial),
            ),
        ]
    )

    # add extensions
    extensions = [crypto_x509.Extension(
        crypto_x509.KeyUsage.oid,
        True,
        crypto_x509.KeyUsage(
            digital_signature=True,
            content_commitment=False,
            key_encipherment=True,
            data_encipherment=True,
            key_agreement=False,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False,
        )
    )]
    if limited:
        proxyinfoext = crypto_x509.UnrecognizedExtension(
            PROXY_CERT_INFO_EXT_OID,
            # language:1.3.6.1.4.1.3536.1.1.1.9
            b"0\x0f0\r\x06\x0b+\x06\x01\x04\x01\x9bP\x01\x01\x01\t",
        )
    else:
        proxyinfoext = crypto_x509.UnrecognizedExtension(
            PROXY_CERT_INFO_EXT_OID,
            # language:Inherit all
            b"0\x0c0\n\x06\x08+\x06\x01\x05\x05\x07\x15\x01",
        )

    # build self-signed proxy certificate
    expiry = _not_valid_after(cert)
    now = datetime.datetime.now(datetime.timezone.utc)
    builder = crypto_x509.CertificateBuilder(
        issuer_name=cert.subject,
        subject_name=proxy_subject,
        public_key=proxy_public_key,
        serial_number=serial,
        not_valid_before=_not_valid_before(cert, naive=True),
        not_valid_after=min(now + datetime.timedelta(hours=minhours), expiry),
        extensions=extensions,
    ).add_extension(proxyinfoext, critical=True)
    proxy = builder.sign(
        private_key=key,
        algorithm=hashes.SHA256(),
        backend=default_backend(),
    )

    return proxy, proxy_private_key



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Common tests for ciecplib entry points.

Inspired by the test_help module from ligo.skymap.tool.tests,
thanks Leo!
"""

import subprocess

import pytest

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # python < 3.8
    importlib_metadata = pytest.importorskip("importlib_metadata")

from ... import (__version__, __name__)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

try:
    ENTRY_POINTS = [
        ep
        for ep in importlib_metadata.distribution(__name__).entry_points
        if ep.group == "console_scripts"
    ]
except importlib_metadata.PackageNotFoundError:
    ENTRY_POINTS = []
PARAMETRIZE_ENTRY_POINTS = pytest.mark.parametrize(
    "ep", [
        pytest.param(ep, id=ep.name)
        for ep in ENTRY_POINTS
    ],
)


def _test_entry_point(ep, args):
    main = ep.load()
    with pytest.raises(SystemExit) as exc:
        main(args)
    if exc.value.code != 0:
        raise subprocess.CalledProcessError(exc.value.code, [ep.name, *args])


@PARAMETRIZE_ENTRY_POINTS
def test_help(ep):
    return _test_entry_point(ep, ["--help"])


@PARAMETRIZE_ENTRY_POINTS
def test_version(capsys, ep):
    _test_entry_point(ep, ["--version"])
    out, err = capsys.readouterr()
    assert out.strip() == __version__



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`ciecplib.tool.utils`."""

import argparse
from unittest import mock

import pytest

from ... import __version__ as ciecplib_version
from ...utils import EcpIdentityProvider
from .. import utils as tools_utils

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def test_help_formatter():
    parser = argparse.ArgumentParser(
        formatter_class=tools_utils.HelpFormatter,
    )
    parser.add_argument("test")
    help_ = parser.format_help()
    assert "Usage: " in help_


def test_argument_parser(capsys):
    parser = tools_utils.ArgumentParser()

    # check that HelpFormatter is in place
    assert parser.formatter_class is tools_utils.HelpFormatter

    # check that the version gets parsed
    with pytest.raises(SystemExit):
        parser.parse_args(["--version"])
    assert capsys.readouterr()[0].strip() == ciecplib_version


@mock.patch("ciecplib.tool.utils.find_principal", side_effect=RuntimeError)
@mock.patch.dict("os.environ", clear=True)
def test_argument_parser_kerberos_error(capsys):
    parser = tools_utils.ArgumentParser()
    with pytest.raises(SystemExit):
        parser.parse_args(["--kerberos"])


@mock.patch("ciecplib.tool.utils.find_principal", return_value="user@REALM")
@mock.patch.dict("os.environ", clear=True)
def test_argument_parser_kerberos_defaults(capsys):
    parser = tools_utils.ArgumentParser()
    args = parser.parse_args(["--kerberos"])
    assert args.username == "user"
    assert args.identity_provider == "REALM"


@mock.patch(
    "ciecplib.tool.utils.get_idps",
    return_value=[
        EcpIdentityProvider("Inst 1", "https://url1", False),
        EcpIdentityProvider("Cardiff University", "https://cardiff", True),
    ],
)
def test_list_idps_action(_, capsys):
    parser = tools_utils.ArgumentParser()
    with pytest.raises(SystemExit):
        parser.parse_args(["--list-idps"])
    stdout = capsys.readouterr()[0]
    assert "'Cardiff University'" in stdout



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""ECP-integated requests session."""

from functools import wraps

from requests_ecp import Session as ECPSession

from .cookies import ECPCookieJar
from .env import _get_default_idp
from .kerberos import (
    has_credential as has_krb5_credential,
    find_principal as find_krb5_principal,
    realm as krb5_realm,
)
from .logging import (
    init_logging,
    reset_logging,
)
from .utils import get_idp_url

__all__ = [
    "Session",
]


class Session(ECPSession):
    """`requests.Session` with default ECP auth and pre-populated cookies."""

    def __init__(
            self,
            idp=_get_default_idp(),
            kerberos=None,
            username=None,
            password=None,
            cookiejar=None,
            debug=False,
            **kwargs,
    ):

        if kerberos is None:
            kerberos = has_krb5_credential()
        if kerberos and not idp:
            idp = krb5_realm(find_krb5_principal())
        if not kerberos and not idp:
            raise ValueError(
                "no Identity Provider (IdP) given, and no kerberos "
                "credential discovered, unable to dynamically determine IdP "
                "endpoint",
            )

        # open session with ECP authentication
        super().__init__(
            idp=get_idp_url(idp) if idp else None,
            kerberos=kerberos,
            username=username,
            password=password,
            **kwargs,
        )

        # load cookies from existing jar or file
        self.cookies = ECPCookieJar()
        if cookiejar:
            self.cookies.update(cookiejar)

        # configure debugging
        self.debug = debug
        if self.debug:
            init_logging(level="DEBUG")

    @wraps(ECPSession.close)
    def close(self):
        if self.debug:
            reset_logging()
        return super().close()



# Copyright (C) 2022-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Logging utilities for CIECPLib."""

import logging
from http.client import HTTPConnection

logging.basicConfig()
ROOT_LOGGER = logging.getLogger()
URLLIB3_LOGGER = logging.getLogger("requests.packages.urllib3")
LOGGERS = (
    ROOT_LOGGER,
    URLLIB3_LOGGER,
)
_DEFAULT_LEVEL = {logger: logger.level for logger in LOGGERS}


def init_logging(level=logging.DEBUG):
    """Enable logging in requests.

    This function is taken from
    https://requests.readthedocs.io/en/v2.9.1/api/#api-changes
    """
    if level is None:  # restore to default
        level = _DEFAULT_LEVEL
    else:
        level = {logger: level for logger in LOGGERS}
    # set level
    for logger in LOGGERS:
        logger.setLevel(level[logger])
    # set debug logging for connections
    debug = ROOT_LOGGER.isEnabledFor(logging.DEBUG)
    HTTPConnection.debuglevel = int(debug)
    URLLIB3_LOGGER.propagate = debug
    return URLLIB3_LOGGER


def reset_logging():
    """Reset the logging levels back to their original values."""
    return init_logging(level=None)



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Environment defaults for ciecplib.

This module parses the following environment variables to set default
settings for CIECP tools:

``ECP_IDP``
   the name or URL of the default ECP Identity Provider (IdP)

Defaults may also be parsed from the ``CIGETCERTOPTS`` environment
variable to support legacy users of ``cigetcert``.
"""

import argparse
import os

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _parse_cigetcertops():
    """Parse the ``CIGETCERTOPTS`` variable for known options.

    Returns
    -------
    opts : `dict`
        a dict of options parsed from the variable
    """
    try:
        opts = os.environ["CIGETCERTOPTS"]
    except KeyError:
        return {}
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--institution")
    args, _ = parser.parse_known_args(opts.split())
    return vars(args)


def _get_default_idp():
    """Return the default ECP IdP, or `None` if no default is found."""
    try:
        return os.environ["ECP_IDP"]
    except KeyError:
        return _parse_cigetcertops().get("institution")


DEFAULT_IDP = _get_default_idp()



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import pytest

from .. import kerberos as ciecplib_kerberos

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize(("lifetime", "result"), [
    (None, False),
    (0, False),
    (1, True),
])
@mock.patch("gssapi.Credentials")
def test_has_credential(creds, lifetime, result):
    # check we can import things
    gssapi = pytest.importorskip("gssapi")
    pytest.importorskip("requests_gssapi")

    # mock the credential
    creds.return_value.name = gssapi.Name(
        "marie.curie@LIGO.ORG",
        name_type=gssapi.NameType.kerberos_principal,
    )
    creds.return_value.lifetime = lifetime

    # assert the result
    assert ciecplib_kerberos.has_credential() is result


@mock.patch.dict("sys.modules", {"gssapi": None})
def test_has_credential_no_gssapi():
    assert not ciecplib_kerberos.has_credential()


@mock.patch("requests_ecp.auth._import_kerberos_auth", side_effect=ImportError)
def test_has_credential_no_kerberos_auth(_):
    assert not ciecplib_kerberos.has_credential()


@mock.patch("gssapi.Credentials")
def test_find_principal(creds):
    gssapi = pytest.importorskip("gssapi")
    # mock the credential
    creds.return_value.name = gssapi.Name(
        "marie.curie@LIGO.ORG",
        name_type=gssapi.NameType.kerberos_principal,
    )
    assert ciecplib_kerberos.find_principal() == "marie.curie@LIGO.ORG"


@mock.patch.dict("sys.modules", {"gssapi": None})
def test_find_principal_no_gssapi():
    with pytest.raises(RuntimeError):
        assert ciecplib_kerberos.find_principal()


@mock.patch("gssapi.Credentials")
def test_find_principal_no_creds(creds):
    gssapi = pytest.importorskip("gssapi")
    creds.side_effect = gssapi.exceptions.GSSError(0, 0)
    with pytest.raises(RuntimeError):
        assert ciecplib_kerberos.find_principal()


def test_realm():
    assert ciecplib_kerberos.realm("marie.curie@EXAMPLE.ORG") == "EXAMPLE.ORG"


@pytest.mark.parametrize("principal", [
    "marie.curie",
    "test/robot",
    "",
])
def test_realm_error(principal):
    with pytest.raises(IndexError):
        ciecplib_kerberos.realm(principal)



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

r"""Authenticate and store session cookies.

ecp-get-cookie queries a SAML/ECP-enabled service, automatically performing
authentication where required, and saves cookies to use in future requests
to the same service.

There are two usages:

    $ ecp-get-cookie -i 'My Institution' -u jsmith https://campus01.edu/my/secret/page

to authenticate with a password prompt, or

    $ ecp-get-cookie -i 'My Institution' -k https://campus01.edu/my/secret/page

to reuse an existing kerberos (``kinit``) credential.

By default the cookie file is created and stored in a location
defined by either

- ``/tmp/ecpcookie.u{uid}`` (Unix), or
- ``C:\Windows\Temp\ecpcookie.{username}`` (Windows)
"""  # noqa: E501

import sys

from ..sessions import Session
from ..cookies import (
    ECPCookieJar,
    has_session_cookies,
    load_cookiejar,
)
from ..ui import get_cookie
from ..utils import DEFAULT_COOKIE_FILE
from .utils import (
    ArgumentParser,
    destroy_file,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def create_parser():
    """Create a command-line argument parser.

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        prog="ecp-get-cookie",
        add_auth=True,
        add_helpers=True,
    )
    parser.add_argument(
        "target_url",
        metavar="URL",
        help="service URL for which to generate cookies",
    )
    parser.add_argument(
        "-c",
        "--cookiefile",
        metavar="cookiefile",
        default=DEFAULT_COOKIE_FILE,
        help="cookie file to create/reuse/destroy",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="write debug output to stdout (implies --verbose)",
    )
    parser.add_argument(
        "-r",
        "--reuse",
        default=False,
        action="store_true",
        help="reuse existing cookies if possible",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="write verbose output to stdout",
    )
    parser.add_argument(
        "-X",
        "--destroy",
        action="store_true",
        default=False,
        help="destroy existing cookie file"
    )
    return parser


def parse_args(parser, args=None):
    """Parse and validate the command-line arguments.

    Returns
    -------
    args : `argparse.Namespace`
    """
    # if user has given -X/--destroy, and isn't just asking for help
    # allow the URL positional argument to be empty
    sargs = set(sys.argv[1:] if args is None else args)
    if {"-X", "--destroy"} & sargs and not {"-h", "--help"} & sargs:
        parser._get_positional_actions()[0].nargs = "?"

    args = parser.parse_args(args=args)

    if args.debug:
        args.verbose = True

    return args


def main(args=None):
    parser = create_parser()
    args = parse_args(parser, args=args)

    def vprint(*pargs, **kwargs):
        """Execute `print` only if --verbose was given."""
        if args.verbose:
            print(*pargs, **kwargs)

    # if asked to destroy, just do that
    if args.destroy:
        destroy_file(args.cookiefile, "cookie file", verbose=args.verbose)
        return 0

    # load old cookies (erroring if file is malformed only)
    try:
        cookiejar = load_cookiejar(args.cookiefile, strict=True)
    except FileNotFoundError:
        cookiejar = ECPCookieJar()

    # if we can't or won't reuse cookies, get a new one
    if not args.reuse or not has_session_cookies(
            cookiejar,
            args.target_url,
    ):
        vprint("Initialising new session...")
        with Session(
            idp=args.identity_provider,
            cookiejar=cookiejar,
            username=getattr(args, "username", None),
            kerberos=args.kerberos,
        ) as sess:
            get_cookie(args.target_url, session=sess)

            # write cookies back to the file
            vprint("Storing cookies...")
            sess.cookies.save(
                args.cookiefile,
                ignore_discard=True,
                ignore_expires=True,
            )
        vprint("Cookie stored in '{0!s}'".format(args.cookiefile))
    else:
        vprint("Reusing existing cookies")

    # load the cert from file to print information
    if args.verbose:
        # reload cookies from the jar,
        # see https://github.com/duncanmmacleod/ciecplib/issues/109
        cookiejar = load_cookiejar(args.cookiefile, strict=True)
        # print information on each cookie
        info = [(cookie.domain, cookie.path, cookie.secure, cookie.name) for
                cookie in cookiejar]
        fmt = "%-{0}s %-5s %-7s %s".format(max(len(i[0]) for i in info))
        print(fmt % ("------", "----", "-------", "----"))
        print(fmt % ("DOMAIN", "PATH", r"SECURE?", "NAME"))
        print(fmt % ("------", "----", "-------", "----"))
        for tup in info:
            print(fmt % tup)
        print(fmt % ("------", "----", "-------", "----"))


if __name__ == "__main__":
    main()



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`ciecplib.tool.ecp_cert_info`."""

from unittest import mock

import pytest

from .. import ecp_cert_info
from ...x509 import _x509_name_str as x509_name_str

MOD_PATH = ecp_cert_info.__name__


@mock.patch(f"{ecp_cert_info.__name__}.load_cert", mock.Mock())
@mock.patch(f"{ecp_cert_info.__name__}.print_cert_info", mock.Mock())
@mock.patch(
    f"{ecp_cert_info.__name__}.time_left",
    mock.Mock(side_effect=(3700., 3500.)),
)
def test_valid():
    """Check that the --valid option for ecp-cert-info works."""
    ecp_cert_info.main(["--valid", "1:0"])
    with pytest.raises(ValueError):
        ecp_cert_info.main(["--valid", "1:0"])


@mock.patch(f"{ecp_cert_info.__name__}.load_cert")
def test_main(load_cert, capsys, x509):
    load_cert.return_value = x509
    ecp_cert_info.main([])
    out = capsys.readouterr().out
    assert out.startswith("subject  : /CN=albert einstein/C=UK")


@mock.patch(f"{ecp_cert_info.__name__}.load_cert", mock.Mock())
@mock.patch(f"{ecp_cert_info.__name__}.time_left")
@pytest.mark.parametrize(("remaining", "code"), [
    (3700, 0),
    (3500, 1),
])
def test_exists(time_left, remaining, code):
    time_left.return_value = remaining
    assert ecp_cert_info.main(["--exists", "--valid", "1:0"]) == code


@mock.patch(f"{ecp_cert_info.__name__}.load_cert")
def test_display(load_cert, x509, capsys):
    """Check that the display option works."""
    load_cert.return_value = x509
    ecp_cert_info.main(["-subject"])
    out = capsys.readouterr().out
    assert out.strip() == x509_name_str(x509.subject)



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Common utilities for tools."""

import argparse
import os
import sys
from functools import wraps
from operator import attrgetter

from .. import __version__
from ..cookies import (
    has_session_cookies,
)
from ..env import _get_default_idp
from ..kerberos import find_principal
from ..utils import get_idps

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _parse_kerberos_principal():
    principal = find_principal()
    return principal.split("@", 1)


# -- argparse helpers ---------------------------

class HelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
):
    def _format_usage(self, usage, actions, groups, prefix):
        if prefix is None:
            prefix = "Usage: "
        return super()._format_usage(
            usage,
            actions,
            groups,
            prefix,
        )


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        add_auth = kwargs.pop("add_auth", True)
        add_helpers = kwargs.pop("add_helpers", True)
        version = kwargs.pop("version", __version__)
        manpage = kwargs.pop("manpage", None)
        mandesc = kwargs.pop("man_short_description", None)

        kwargs.setdefault("formatter_class", HelpFormatter)
        kwargs.setdefault("add_help", not add_helpers)
        super(ArgumentParser, self).__init__(*args, **kwargs)

        self._positionals.title = "Required arguments"
        if add_auth:
            self._optionals.title = "Other options"
        else:
            self._optionals.title = "Options"

        # add manpage options for argparse-manpage
        self._manpage = manpage
        if not mandesc and self.description:
            # if man_short_description was not given, copy the first line of
            # the parser description, with some pedantic reformatting
            mandesc = (
                self.description.split("\n")[0].lower().rstrip(".")
            )
        self.man_short_description = mandesc

        # add auth options group
        if add_auth:
            self.add_auth_arguments()
            group = self._action_groups.pop(-1)
            self._action_groups.insert(1, group)

        # add helper commands group
        if add_helpers:
            self.add_helper_arguments(version=version)

    def add_helper_arguments(self, version=__version__):
        helpers = self.add_argument_group("Helper arguments")
        helpers.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="show this help message and exit",
        )
        helpers.add_argument(
            "-l",
            "--list-idps",
            action=ListIdpsAction,
        )
        if version is not None:
            helpers.add_argument(
                "-V",
                "--version",
                action="version",
                version=version,
            )
        return helpers

    def add_auth_arguments(
            self,
            identity_provider=True,
            kerberos=True,
            username=True,
    ):
        auth = self.add_argument_group("Authentication options")

        if identity_provider:
            defaultidp = _get_default_idp()
            auth.add_argument(
                "-i",
                "--identity-provider",
                default=defaultidp,
                help="name of institution providing the identity (e.g. "
                     "'Cardiff University'), or domain name of IdP host "
                     "(e.g. idp.cf.ac.uk), see --list-idps for a list of "
                     "Identity Provider (IdPs) and their IdP URL. Shortened "
                     "institution names (e.g. 'Cardiff') can be given as "
                     "long as they uniquely match a full institution name "
                     "known by CILogon",
            )

        if kerberos:
            auth.add_argument(
                "-k",
                "--kerberos",
                action="store_true",
                default=False,
                help="enable kerberos negotiation",
            )

        if username:
            auth.add_argument(
                "-u",
                "--username",
                help="authentication username, will be prompted for if not "
                     "given and not using --kerberos",
            )

        return auth

    @wraps(argparse.ArgumentParser.parse_args)
    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)
        # if -k/--kerberos was given, try and use it to set defaults
        if getattr(args, "kerberos", None):
            self._set_defaults_from_kerberos_principal(args)
        # if -X/--destroy wasn't given and
        # -i/--identity-provider also wasn't given (and is supported)
        # then raise an error
        #    - this just supports giving -X/--destroy without having to
        #      also give -i/--identity-provider for no reason
        destroy = getattr(args, "destroy", None)
        idp = getattr(args, "identity_provider", False)
        if not destroy and idp is None:
            self.error(
                "the following arguments are required: -i/--identity-provider",
            )
        return args

    @staticmethod
    def _set_defaults_from_kerberos_principal(args):
        """Set defaults for username and IdP based on a kerberos principal."""
        try:
            username, realm = _parse_kerberos_principal()
        except Exception:
            return
        for arg, default in [
            ("username", username),
            ("identity_provider", realm),
        ]:
            if getattr(args, arg, "_notset_") is None:
                setattr(args, arg, default)


class ListIdpsAction(argparse.Action):
    def __init__(
            self,
            option_strings,
            dest=argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            help="show IdP names and URLs and exit",
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        idps = get_idps()
        formatter = parser._get_formatter()
        fmt = "{0.name!r:%ss} : {0.url}" % (max(len(i.name) for i in idps) + 2)
        lines = []
        for idp in sorted(idps, key=attrgetter("name")):
            lines.append(fmt.format(idp))
        formatter.add_text("\n".join(lines))
        parser._print_message(formatter.format_help(), sys.stdout)
        parser.exit()


# -- miscellaneous ------------------------------

def reuse_cookies(cookiejar, url, verbose=False):
    """Determine if a cookiejar has session cookies we can reuse.

    Parameters
    ----------
    cookiefile : `str`
        the path to the cookie file

    url : `str`
        the URL of the service being accessed with cookies

    verbose : `bool`, optional
        if `True`, print verbose output

    Returns
    -------
    cookiejar : `ligo.org.cookies.ECPCookieJar`, `NoneType`
        the cookie jar (if the cookie file was read correctly), otherwise
        `None`
    reuse : `bool`
        `True` if the cookies in the jar can be reused to access the given
        service URL, otherwise `False`
    """
    if verbose:
        print("Validating existing cookies...", end=" ")
    reuse = has_session_cookies(cookiejar, url)
    if verbose and reuse:
        print("OK")
    elif verbose:
        print("cannot reuse, need new cookies")
    return cookiejar, reuse


def destroy_file(path, desc=None, verbose=False):
    """Destroy a file (if it exists), with verbose output."""
    if verbose:
        desc = (desc or "").rstrip(" ") + " "
        print("Removing {}'{!s}'...".format(desc, path), end=" ")
    try:
        os.unlink(path)
    except FileNotFoundError:  # file doesn't exit, no matter
        if verbose:
            print("not found", end="")
    else:
        if verbose:
            print("done", end="")
    finally:
        if verbose:
            print()



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Common fixtures for ciecplib tests."""

import datetime

from cryptography import x509 as crypto_x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import Encoding

import pytest


@pytest.fixture(scope="session")  # one per suite is fine
def private_key():
    """Create an RSA private key for testing."""
    return generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )


@pytest.fixture(scope="session")
def public_key(private_key):
    """Return the ``public_key()`` for ``private_key``."""
    return private_key.public_key()


@pytest.fixture
def x509(public_key, private_key):
    """Create a `cryptography.x509.Certificate` for testing."""
    name = crypto_x509.Name([
        crypto_x509.NameAttribute(typ, value) for (typ, value) in (
            (crypto_x509.NameOID.COMMON_NAME, "albert einstein"),
            (crypto_x509.NameOID.COUNTRY_NAME, "UK"),
            (crypto_x509.NameOID.STATE_OR_PROVINCE_NAME, "Wales"),
            (crypto_x509.NameOID.LOCALITY_NAME, "Cardiff"),
            (crypto_x509.NameOID.ORGANIZATION_NAME, "Cardiff University"),
            (crypto_x509.NameOID.ORGANIZATIONAL_UNIT_NAME, "Gravity"),
        )
    ])
    now = datetime.datetime.now(datetime.timezone.utc)
    return crypto_x509.CertificateBuilder(
        issuer_name=name,
        subject_name=name,
        public_key=public_key,
        serial_number=1000,
        not_valid_before=now,
        not_valid_after=now + datetime.timedelta(seconds=86400),
    ).sign(private_key, hashes.SHA256(), backend=default_backend())


@pytest.fixture
def x509_path(tmp_path, x509):
    """Write ``x509`` to a temporary file in PEM format, and return the path."""
    path = tmp_path / "tmpx509.pem"
    with path.open("wb") as tmp:
        tmp.write(x509.public_bytes(Encoding.PEM))
    return path



# Copyright (C) 2020-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for :mod:`ciecplib.cookies`."""

from http.cookiejar import Cookie
try:
    from http.cookiejar import NETSCAPE_HEADER_TEXT
except ImportError:  # python < 3.10
    from http.cookiejar import MozillaCookieJar
    NETSCAPE_HEADER_TEXT = MozillaCookieJar.header

import pytest

from ciecplib import cookies as ciecplib_cookies


@pytest.fixture
def sessioncookie():
    return Cookie(
        version=0,
        name="_shibsession_1234567890",
        value="_1234567890",
        port=None,
        port_specified=False,
        domain="somewhere.test.com",
        domain_specified=False,
        domain_initial_dot=False,
        path="/",
        path_specified=True,
        secure=True,
        expires=None,
        discard=True,
        comment=None,
        comment_url=None,
        rest={"HttpOnly": None},
        rfc2109=False,
    )


@pytest.fixture
def ecpcookiejar(sessioncookie):
    jar = ciecplib_cookies.ECPCookieJar()
    jar.set_cookie(sessioncookie)
    return jar


class TestECPCookieJar(object):
    TEST_CLASS = ciecplib_cookies.ECPCookieJar

    def test_save_discard(self, ecpcookiejar, tmp_path):
        path = tmp_path / "cookies"
        ecpcookiejar.save(path)
        # check that the header got written
        assert path.read_text().startswith(NETSCAPE_HEADER_TEXT)
        # check that there aren't actually any cookies
        # (since the cookie is a session cookie [`discard=True`])
        jar = self.TEST_CLASS()
        jar.load(str(path))
        assert len(jar) == 0

    def test_save_ignore_discard(self, ecpcookiejar, tmp_path):
        path = tmp_path / "cookies"
        ecpcookiejar.save(path, ignore_discard=True, ignore_expires=True)
        jar = self.TEST_CLASS()
        jar.load(str(path), ignore_discard=True, ignore_expires=True)
        assert jar["_shibsession_1234567890"] == "_1234567890"


def test_extract_session_cookie(ecpcookiejar, sessioncookie):
    assert ciecplib_cookies.extract_session_cookie(
        ecpcookiejar,
        "https://somewhere.test.com/blah/blah",
    ) is sessioncookie


def test_extract_session_cookie_error(ecpcookiejar):
    with pytest.raises(ValueError):
        ciecplib_cookies.extract_session_cookie(
            ecpcookiejar,
            "https://somethingelse.test.com",
        )


@pytest.mark.parametrize("url, result", [
    ("https://somewhere.test.com", True),
    ("https://somewhereelse.test.com", False),
])
def test_has_session_cookies(ecpcookiejar, url, result):
    assert ciecplib_cookies.has_session_cookies(ecpcookiejar, url) is result


def test_load_cookiejar(ecpcookiejar, tmp_path):
    path = tmp_path / "cookies"
    ecpcookiejar.save(path, ignore_discard=True, ignore_expires=True)
    jar = ciecplib_cookies.load_cookiejar(path)
    assert jar == ecpcookiejar


def test_load_cookiejar_error(tmp_path):
    with pytest.raises(OSError):
        ciecplib_cookies.load_cookiejar(tmp_path / "blah")
    assert isinstance(
        ciecplib_cookies.load_cookiejar(tmp_path / "blah", strict=False),
        ciecplib_cookies.ECPCookieJar,
    )



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Kerberos utilities for ciecplib."""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _get_creds(usage="initiate", **kwargs):
    """Return the active GSSAPI credentials, if any."""
    try:
        import gssapi
    except ImportError:
        return None
    try:
        return gssapi.Credentials(usage=usage, **kwargs)
    except gssapi.exceptions.GSSError:
        return None


def has_credential(**store_kw):
    """Return `True` if an active Kerberos (GSSAPI) credential is available.

    Parameters
    ----------
    store_kw
        Keyword arguments to pass to the credentials store extension of
        the underlying GSSAPI implementation.
        See https://pythongssapi.github.io/python-gssapi/credstore.html
        for more details.

    Returns
    -------
    True
        If a valid Kerberos credential is found with a lifetime of
        more than 1 second.
    False
        Otherwise.

    Examples
    --------
    >>> has_credential()
    False
    >>> has_credential(
    ...     ccache="mykrb5cc",
    ...     client_keytab="/home/user/.kerberos/me.keytab",
    ... )
    True

    Notes
    -----
    This function will always return `False` if the requests-gssapi
    Kerberos Auth plugin required by requests-ecp is not found.

    See Also
    --------
    gssapi.Credentials
        For more details on how credentials are accessed using
        Python-GSSAPI.
    """
    try:
        from requests_ecp.auth import _import_kerberos_auth
    except ImportError:
        # probably requests-ecp < 0.3.0, which always supports kerberos
        pass
    else:
        try:
            _import_kerberos_auth()
        except ImportError:
            # requests-ecp doesn't support Kerberos, so we can't
            return False

    # get credentials and inspect to see if they have a valid liftime
    creds = _get_creds()
    return creds is not None and (creds.lifetime or 0) >= 1


def find_principal():
    """Determine the principal for an active kerberos credential.

    Returns
    -------
    name: `str`
        The `str` representation of the Kerberos principal.

    Raises
    ------
    RuntimeError
        If no credential is found.
    """
    creds = _get_creds()
    if creds is None:
        raise RuntimeError(
            "failed to find active GSSAPI (Kerberos) credential",
        )
    return str(creds.name)


def realm(principal):
    """Return the kerberos realm name from a principal.

    Parameters
    ----------
    principal : `str`
        the kerberos principal to parse

    Returns
    -------
    realm : `str`
        the realm name

    Raises
    ------
    IndexError
        If the input principal cannot be parsed properly
        (normally if it doesn't contain an ``@`` character).

    Examples
    --------
    >>> realm('marie.curie@EXAMPLE.ORG')
    'EXAMPLE.ORG'
    """
    return principal.rsplit("@", 1)[1]



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

import os
import random
import re
import string
from collections import namedtuple
from pathlib import Path

import requests

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- default paths ------------------------------------------------------------

def _tmpfile(prefix):
    if os.name == "nt":
        tmpdir = Path(os.environ["SYSTEMROOT"]) / "Temp"
        user = os.getlogin()
    else:
        tmpdir = "/tmp"
        user = "u{}".format(os.getuid())
    return Path(tmpdir) / "{}{}".format(prefix, user)


def get_ecpcookie_path():
    """Return the default path for the ECP cookie file.

    Returns
    -------
    path : `pathlib.Path`
    """
    return _tmpfile("ecpcookie.")


def get_x509_proxy_path():
    """Return the default path for the X.509 certificate file.

    Returns
    -------
    path : `pathlib.Path`
    """
    if os.getenv("X509_USER_PROXY"):
        return Path(os.environ["X509_USER_PROXY"])
    return _tmpfile("x509up_")


DEFAULT_COOKIE_FILE = str(get_ecpcookie_path())
DEFAULT_X509_USER_FILE = str(get_x509_proxy_path())

# -- institution URLs ----------------------------------------------------

DEFAULT_IDPLIST_URL = "https://cilogon.org/include/ecpidps.txt"
DEFAULT_SP_URL = "https://ecp.cilogon.org/secure/getcert"
_ECP_ENDPOINT_REGEX = re.compile(r"\Ahttps://.*/SOAP/ECP\Z")
_KERBEROS_SUFFIX = " (Kerberos)"
_PRIMARY_SUFFIX_REGEX = re.compile(r" (\()?({})(\))?\Z".format("|".join((
    "1",
    "main",
    "primary",
    "principal",
))))
_SECONDARY_SUFFIX_REGEX = re.compile(r" (\()?({})(\))?\Z".format("|".join((
    "[2-9]",
    "[0-9][0-9]+",
    "backup",
    "secondary",
    "test",
))))
_URL_REGEX = re.compile(r"[^ \t\n\r\f\vA-Z]+")


EcpIdentityProvider = namedtuple(
    "EcpIdentityProvider",
    (
        "name",
        "url",
        "iskerberos",
    ),
)


def get_idps(url=DEFAULT_IDPLIST_URL, timeout=10):
    """Download the list of known ECP IdPs from the given URL.

    The output is a `list` of `EcpIdentityProvider` objects.

    Some institutions may have two entries if they also support Kerberos.

    Parameters
    ----------
    url : `str`
        the URL of the IDP list file
    """
    idps = list()
    for line in requests.get(
        url,
        timeout=timeout,
        stream=True,
    ).iter_lines():
        url, inst = line.decode("utf-8").strip().split(" ", 1)
        idps.append(EcpIdentityProvider(
            inst,
            url,
            inst.endswith(_KERBEROS_SUFFIX),
        ))
    return idps


def _match(value, idplist, attr, kerberos=None):
    return [
        inst for inst in idplist
        if value in getattr(inst, attr).lower()
        and kerberos in (None, inst.iskerberos)
    ]


def _preferred_match(matches):
    """Attempt to select a preferred match from a list of many.

    This only prefers institutions names that don't end with a
    numeric suffix (or end with ' 1'), mainly to distinguish
    between multiple entries like ``'Cardiff University 1'`` and
    ``'Cardiff University 2'``
    """
    if not len(matches) > 1:
        return matches
    # find all names marked as 'primary' or similar
    primaries = [x for x in matches if _PRIMARY_SUFFIX_REGEX.search(x.name)]
    # find all names marked as 'backup' or similar
    secondaries = [x for x in matches if
                   _SECONDARY_SUFFIX_REGEX.search(x.name)]
    # if only one primary, return that
    if len(primaries) == 1:
        return primaries[:1]
    # if all secondary, bar one, return that
    if len(secondaries) == len(matches) - 1:
        return [(set(matches) - set(secondaries)).pop()]
    # otherwise no preference, return everything
    return matches


def _match_institution(value, institutions, kerberos=None):
    value = str(value).lower()

    # try and match the institution name
    matches = _preferred_match(
        _match(value, institutions, "name", kerberos=kerberos),
    )
    if len(matches) == 1:
        return matches.pop()

    # otherwise match the IdP URL
    if _URL_REGEX.match(value):
        umatches = _preferred_match(
            _match(value, institutions, "url", kerberos=kerberos),
        )
        if len(umatches) == 1:
            return umatches.pop()
        matches = matches or umatches or []

    # if we found multiple matches, print them to help the user
    if len(matches):
        raise ValueError(
            "failed to identify unique IdP URL for {0!r}, possible matches "
            "include:\n"
            "{1}".format(
                value,
                "\n".join(map("{0.name!r}: {0.url}".format, matches)),
            ),
        )

    # otherwise just fail
    raise ValueError("failed to identify IdP URLs for {0!r}".format(value))


def get_idp_url(url_or_name, idplist_url=DEFAULT_IDPLIST_URL, kerberos=False):
    """Return the unique IdP URL for a given institution or URL stub.

    Parameters
    ----------
    url_or_name: `str`
        the name of an institution, or a URL for the endpoint, or part thereof

    idplist_url : `str`, optional
        the URL to query for the list of enabled ECP IdPs

    kerberos : `bool`, optional
        if ``True`` return a Kerberos URL, if available, otherwise return
        a standard SAML/ECP endpoint URL

    Returns
    -------
    url : `str`
        the formatted URL of the IdP ECP endpoint

    Raises
    ------
    ValueError
        if there isn't a unique match for either the institution name, or the
        IdP URL

    Examples
    --------
    >>> get_idp_url("LIGO")
    'https://login.ligo.org/idp/profile/SAML2/SOAP/ECP'
    >>> get_idp_url("ligo.org")
    'https://login.ligo.org/idp/profile/SAML2/SOAP/ECP'
    """
    # short circuit full URLs
    if _ECP_ENDPOINT_REGEX.match(url_or_name):
        return url_or_name
    # otherwise match against CILogon's registered list
    idps = get_idps(url=idplist_url)
    institution = _match_institution(url_or_name, idps, kerberos=kerberos)
    return institution.url


# -- misc utilities -----------------------------------------------------------

def random_string(length, outof=string.ascii_lowercase + string.digits):
    # http://stackoverflow.com/a/23728630/2213647 says SystemRandom()
    # is most secure
    return "".join(random.SystemRandom().choice(outof) for _ in range(length))



# Copyright (C) 2022-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`ciecplib.tool.ecp_get_curl`."""

import pytest

from requests import RequestException

from .. import ecp_curl
from ...cookies import load_cookiejar


def test_main_stdout(capsys, requests_mock):
    """Test ``ecp-curl`` with default options (writes to stdout)."""
    requests_mock.get(
        "https://test.example.com",
        content=b"hello world",
    )
    # curl the github API top-level URL
    ecp_curl.main([
        "https://test.example.com",
        "--identity-provider", "https://test.example.com/SOAP/ECP",
    ])
    out, err = capsys.readouterr()
    # check that we got valid JSON back
    assert out == "hello world"


def test_main_file(requests_mock, tmp_path):
    """Test ``ecp-curl`` writing to file."""
    outfile = tmp_path / "api.github.com.json"
    requests_mock.get(
        "https://test.example.com",
        content=b"hello world",
    )
    # curl the github API top-level URL
    ecp_curl.main([
        "https://test.example.com",
        "--output", str(outfile),
        "--identity-provider", "https://test.example.com/SOAP/ECP",
    ])
    # check that we got valid JSON back
    assert outfile.read_text() == "hello world"


def test_main_error(requests_mock, tmp_path):
    """Test that ``ecp-curl`` propagates RequestExceptions."""
    requests_mock.get(
        "https://test.example.com",
        status_code=404,
    )
    with pytest.raises(RequestException):
        ecp_curl.main([
            "https://test.example.com",
            "--identity-provider", "https://test.example.com/SOAP/ECP",
        ])


def test_main_store_cookie_file(capsys, requests_mock, tmp_path):
    """Test that ``ecp-curl`` can store session cookies (sort of)."""
    jar = tmp_path / "cookies"
    requests_mock.get(
        "https://test.example.com",
        content=b"hello world",
        cookies={"foo": "bar"},
    )

    # curl the github API top-level URL
    ecp_curl.main([
        "https://test.example.com",
        "--store-session-cookies",
        "--cookiefile", str(jar),
        "--identity-provider", "https://test.example.com/SOAP/ECP",
    ])
    out, err = capsys.readouterr()

    # check that we got valid JSON back
    assert out == "hello world"

    # check that we can read the cookie jar
    # (which probably doesn't have any cookies in it)
    load_cookiejar(jar)


def content_callback(request, context):
    """Return response content based on the request context.

    Just mocks a 401 Unauthorized response if no ``Authorization`` header is
    included.
    """
    if request.headers.get("Authorization"):
        context.status_code = 200
    else:
        context.status_code = 401
        context.reason = "Unauthorized"
    return b"hello world"


@pytest.mark.parametrize(("headers", "status"), [
    pytest.param([], 401, id="no-headers"),
    pytest.param(
        [
            "-H", "Custom:something",
        ],
        401,
        id="custom-header",
    ),
    pytest.param(
        [
            "--header=Custom:something",
            "-H", "Authorization: Bearer 123",
        ],
        200,
        id="auth-header",
    ),
])
def test_main_headers(capsys, requests_mock, headers, status):
    """Test ``ecp-curl`` with default options (writes to stdout)."""
    requests_mock.get(
        "https://test.example.com",
        content=content_callback,
    )
    args = [
        "https://test.example.com",
        "--identity-provider", "https://test.example.com/SOAP/ECP",
        *headers,
    ]
    if status == 200:
        ecp_curl.main(args)
        out, err = capsys.readouterr()
        # check that we got valid JSON back
        assert out == "hello world"
        return
    with pytest.raises(
        RequestException,
        match="401 Client Error: Unauthorized",
    ):
        ecp_curl.main(args)



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Transfer a URL using SAML/ECP authentication.

ecp-curl attempts to use existing cookies for the target domain
and prompts for authorisation information if required.
Persistent cookies acquired during transactions are automatically
recorded in the cookie file, but session cookies are discarded (unless
--store-session-cookies is given).

Currently only HTTP GET requests are supported (patches welcome!).
"""

import sys
from email.parser import HeaderParser
from http.client import HTTPMessage

from ..cookies import load_cookiejar
from ..sessions import Session
from ..utils import DEFAULT_COOKIE_FILE
from .utils import (
    ArgumentParser,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def create_parser():
    """Create a command-line argument parser.

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        prog="ecp-curl",
        add_auth=True,
        add_helpers=True,
    )
    parser.add_argument(
        "url",
        help="the URL to transfer"
    )
    parser.add_argument(
        "-c",
        "--cookiefile",
        metavar="FILE",
        default=DEFAULT_COOKIE_FILE,
        help="cookie file to use",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="write debug output (uses both stderr and stdout)",
    )
    parser.add_argument(
        "-H",
        "--header",
        action="append",
        default=None,
        help="HTTP headers to include in the request",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="write to %(metavar)s instead of stdout",
    )
    parser.add_argument(
        "-s",
        "--store-session-cookies",
        action="store_true",
        default=False,
        help="store session cookies in the cookie file"
    )
    return parser


def format_headers(headers):
    """Format headers received on the command line.

    Each header should be formatted according to
    `RFC5322 <https://datatracker.ietf.org/doc/html/rfc5322.html#section-2.2>`_.

    Parameters
    ----------
    headers : `list` of `str`
        A list of header strings (``"Key: Value"``) to parse.

    Returns
    -------
    headerdict : `dict[str, str]`
        A `dict` of ``(key, value)`` headers ready to pass to `requests.get`.
    """
    if not headers:
        return {}
    headerstr = "\n".join(headers)
    return dict(HeaderParser(HTTPMessage).parsestr(headerstr))


def write(data, path):
    """Write ``data`` to a file path, or to stdout.

    Parameters
    ----------
    data : `bytes`
        The data to write.

    path : `str`, `None`
        The target path to write to, or `None` to write to stdout.
    """
    if path is None:
        file = sys.stdout.buffer
    else:
        file = open(path, "wb")
    try:
        file.write(data)
    finally:
        if path is not None:
            file.close()


def main(args=None):
    parser = create_parser()
    args = parser.parse_args(args=args)
    cookiejar = load_cookiejar(
        args.cookiefile,
        strict=False,
    )
    headers = format_headers(args.header)

    with Session(
        cookiejar=cookiejar,
        idp=args.identity_provider,
        username=args.username,
        kerberos=args.kerberos,
        debug=args.debug,
    ) as sess:
        # GET
        resp = sess.get(
            args.url,
            headers=headers,
        )
        resp.raise_for_status()
        # write
        write(resp.content, args.output)
        # store session cookies
        if args.store_session_cookies:
            sess.cookies.save(
                args.cookiefile,
                ignore_discard=True,
                ignore_expires=True,
            )


if __name__ == "__main__":
    main()



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Print information about an existing X.509 credential.

ecp-cert-info prints information from an existing X.509 credential,
including owner information, certificate type, and time remaining
until expiry.
"""

import argparse
import sys
from datetime import timedelta

from ..x509 import (
    load_cert,
    print_cert_info,
    time_left,
)
from ..utils import DEFAULT_X509_USER_FILE
from .utils import (
    ArgumentParser,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

EPILOG = r"""
Environment:

X509_USER_PROXY:
    the default path for the credential file
"""

MANPAGE = [
    {"heading": "environment",
     "content": """
.TP
.B "X509_USER_PROXY"
The default path for the credential file """,
     },
]


def _hours_minutes(argstring):
    """Parse the --valid argument as hours:minutes.

    Returns the argument as a float in seconds.
    """
    try:
        hrs, minutes = argstring.split(":")
    except TypeError as exc:
        exc.args = ("--valid argument must be in the format H:M",)
        raise
    return float(hrs) * 3600. + float(minutes) * 60.


def create_parser():
    """Create a command-line argument parser

    Returns
    -------
    parser : `argparse.ArgumentParser`
    """
    parser = ArgumentParser(
        description=__doc__,
        epilog=EPILOG,
        manpage=MANPAGE,
        prog="ecp-cert-info",
        add_auth=False,
        add_helpers=True,
    )
    parser.add_argument(
        "-f",
        "--file",
        default=DEFAULT_X509_USER_FILE,
        help="certificate file to read (must exist)",
    )
    parser.add_argument(
        "-v",
        "-valid",
        "--valid",
        metavar="HH:MM",
        type=_hours_minutes,
        help="time requirement for proxy to be valid",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="print full text of certificate",
    )
    parser.add_argument(
        "-e",
        "-exists",
        "--exists",
        action="store_true",
        default=False,
        help="return 0 if valid proxy exists, 1 otherwise",
    )

    dispargs = parser.add_argument_group("Display arguments")
    dispargs.add_argument(
        "-s",
        "-subject",
        "--subject",
        action="append_const",
        const="subject",
        dest="display",
        default=argparse.SUPPRESS,
        help="distinguished name (DN) of subject",
    )
    dispargs.add_argument(
        "-i",
        "-issuer",
        "--issuer",
        action="append_const",
        const="issuer",
        dest="display",
        default=argparse.SUPPRESS,
        help="DN of issuer (certificate signer)",
    )
    dispargs.add_argument(
        "-type",
        "--type",
        action="append_const",
        const="type",
        dest="display",
        default=argparse.SUPPRESS,
        help="type of certificate",
    )
    dispargs.add_argument(
        "-t",
        "-timeleft",
        "--timeleft",
        action="append_const",
        const="timeleft",
        dest="display",
        default=argparse.SUPPRESS,
        help="time left until certificate expires",
    )

    dispargs.add_argument(
        "-path",
        "--path",
        action="append_const",
        const="path",
        dest="display",
        default=argparse.SUPPRESS,
        help="path to certificate file",
    )

    return parser


def parse_args(parser, args=None):
    """Parse and validate the command-line arguments

    Returns
    -------
    args : `argparse.Namespace`
    """
    return parser.parse_args(args=args)


def main(args=None):
    parser = create_parser()
    args = parse_args(parser, args=args)

    # load certificate
    cert = load_cert(args.file)

    # determine time left (if we need it)
    if args.valid or args.exists:
        valid = timedelta(seconds=args.valid or 0)
        remaining = timedelta(seconds=time_left(cert))
        isvalid = remaining >= valid

    if args.exists:
        # if proxy exists and is valid, return 0, otherwise 1
        return int(not isvalid)

    # print certificate information
    print_cert_info(
        cert,
        path=args.file,
        display=getattr(args, "display", None),
        verbose=args.verbose,
        stream=sys.stdout,
    )

    # check credential validity if given
    if args.valid and not isvalid:
        raise ValueError(
            f"timeleft ({remaining}) is less than required ({valid})"
        )


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Cookie handling for SAML ECP authentication."""

from copy import deepcopy
from functools import wraps
from http.cookiejar import (LoadError, MozillaCookieJar)
from urllib.parse import urlparse

from requests.cookies import RequestsCookieJar

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- cookie jar ---------------------------------------------------------------

class ECPCookieJar(RequestsCookieJar, MozillaCookieJar):
    """Custom cookie jar that stores cookies in the cURL format."""

    @wraps(MozillaCookieJar.save)
    def save(self, *args, **kwargs):
        copy = deepcopy(self)
        for cookie in copy:
            if cookie.expires is None:
                # hack the cookie to store 0 so that cURL will use it
                cookie.expires = "0"
        # now save our modified cookiejar
        return super(ECPCookieJar, copy).save(*args, **kwargs)

    def _really_load(self, *args, **kwargs):
        out = super()._really_load(*args, **kwargs)
        for cookie in self:
            if not cookie.expires:  # reformat "0" as None
                cookie.expires = None
        return out


# -- utilities ----------------------------------------------------------------

def extract_session_cookie(jar, url):
    """Return a session cookie for the given URL from the jar.

    Parameters
    ----------
    jar : `http.cookiejar.CookieJar`
        the cookie jar to check

    url : `str`
        the URL of the service that needs cookies

    Returns
    -------
    cookie : `http.cookiejar.Cookie`
        the relevant cookie

    Raises
    ------
    ValueError
        if no appropriate cookie is found
    """
    url = urlparse(url).netloc
    for cookie in list(jar)[::-1]:
        if (
                cookie.name.startswith("_shibsession_")
                and cookie.domain == url
                and cookie.expires is None
        ):
            return cookie
    raise ValueError(
        "no shibsession cookie found for {!r}".format(url),
    )


def has_session_cookies(jar, url):
    """Return `True` if the given cookie jar has a session cookie we can use.

    Parameters
    ----------
    jar : `http.cookiejar.CookieJar`
        the cookie jar to check

    url : `str`
        the URL of the service that needs cookies

    Returns
    -------
    can_reuse : `bool`
        `True` if any cookie in the jar is a non-expiring ``shibsession``
        cookie for the given same domain as ``url``
    """
    try:
        extract_session_cookie(jar, url)
    except ValueError:
        return False
    return True


def load_cookiejar(
        cookiefile,
        strict=True,
        ignore_discard=True,
        ignore_expires=True,
):
    """Load a cookie jar from a file.

    Parameters
    ----------
    cookiefile : `str`
        path to cookie jar file

    strict : `bool`, optional
        if `True` (default), raise all exceptions as they occur,
        if `False` just emit warnings

    ignore_discard, ignore_expires : `bool`, optional
        options to pass to :meth:`http.cookiejar.FileCookieJar.load`,
        both default to `True` in this usage.
    """
    cookiejar = ECPCookieJar()
    try:
        cookiejar.load(
            str(cookiefile),
            ignore_discard=ignore_discard,
            ignore_expires=ignore_expires,
        )
    except (LoadError, OSError):
        if strict:
            raise
    return cookiejar



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""User-interface functions for SAML ECP authentication."""

import math
import string
import warnings
from urllib import parse as urllib_parse

from requests import HTTPError

from .env import DEFAULT_IDP
from .cookies import extract_session_cookie
from .requests import _ecp_session
from .utils import (
    DEFAULT_SP_URL,
    random_string,
)
from .x509 import load_pkcs12

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@_ecp_session
def get_cookie(
        url,
        endpoint=DEFAULT_IDP,
        username=None,
        kerberos=False,
        debug=False,
        session=None,
):
    """Create a SAML/ECP session cookie valid for the given URL.

    Parameters
    ----------
    url : `str`
        the target URL/domain

    endpoint : `str`, optional
        the identity provider URL

    username : `str`, optional
        the username to use when authenticating

    kerberos : `bool`, optional
        if `True` use an existing kerberos TGT to authenticate

    debug : `bool`, optional
        if `True` enable verbose debugging from requests, currently unused

    session : `requests.Session`, optional
        an active `requests.Session` to use with the query

    return_all : `bool`, optional
        return all cookies from the authentication request

    Returns
    -------
    cookie : `http.cookiejar.Cookie`
        the newly-minted session cookie
    """
    # decorator guarantees a populated session
    with session as sess:
        # authenticate against the endpoint
        sess.ecp_authenticate(
            url=url,
        )

        # make the original request, it may introduce more cookies
        if url != DEFAULT_SP_URL:
            resp = sess.get(url=url)
            resp.raise_for_status()

        # extract the shibsession cookie for the SP:
        #   we do this by searching for all cookies associated with
        #   the relevant SP domain, and picking the most recent one
        return extract_session_cookie(sess.cookies, url)


@_ecp_session
def get_cert(
        endpoint=DEFAULT_IDP,
        hours=168,
        username=None,
        kerberos=False,
        spurl=DEFAULT_SP_URL,
        debug=False,
        session=None,
):
    """Create an X.509 credential using SAML/ECP.

    .. warning::

        CILogon has announced the retirement of support for X.509
        certificates, please see https://ca.cilogon.org/retirement
        for details; attempts to request a certificate from cilogon.org
        will result in an error after May 2025, please contact
        help@cilogon.org for assistance.

    Parameters
    ----------
    endpoint : `str`, optional
        the identity provider URL

    hours : `int`, optional
        the desired validity of the credential

    username : `str`, optional
        the username to use when authenticating

    kerberos : `bool`, optional
        if `True` use an existing kerberos TGT to authenticate

    debug : `bool`, optional
        if `True` enable verbose debugging from requests, currently unused

    session : `requests.Session`, optional
        an active `requests.Session` to use with the query

    Returns
    -------
    cert : `cryptography.x509.Certificate`
        The newly minted certificate.

    key : `cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey`
        The RSA key object used to sign the certificate.
    """
    if spurl == DEFAULT_SP_URL:
        warnings.warn(
            "CILogon has announced the retirement of support for X.509 "
            "certificates, please see https://ca.cilogon.org/retirement "
            f"for details; attempts to request a certificate from {spurl} "
            "will result in an error after May 2025, please contact "
            "help@cilogon.org for assistance.",
            DeprecationWarning,
            stacklevel=3,
        )

    # decorator guarantees a populated session
    with session as sess:
        cookie = get_cookie(spurl, session=sess, debug=debug)

        if debug:
            print("Authenticated against IdP with ECP")
            print("Requesting certificate from {0}".format(spurl))

        # request PKCS12 certificate from SP
        csrfstr = random_string(10)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Cookie": "CSRF={0}; {1.name}={1.value}".format(csrfstr, cookie),
        }
        p12password = (
            random_string(16, string.ascii_letters)
            + random_string(2, string.digits)
            + random_string(2, "!@#$%^&*()")
        )
        certdata = urllib_parse.urlencode({
            "submit": "pkcs12",
            "CSRF": csrfstr,
            "p12password": p12password,
            "p12lifetime": math.ceil(hours),
        }).encode("utf-8")
        resp = sess.post(
            spurl,
            data=certdata,
            headers=headers,
        )
        try:
            resp.raise_for_status()
        except HTTPError as exc:
            exc.args = (f"{exc}: '{resp.text}'",)
            raise

        if debug:
            print("Certificate received.")
        return load_pkcs12(
            resp.content,
            p12password.encode("utf-8"),
        )



# Copyright (C) 2019-2025 Cardiff University
#
# This file is part of ciecplib.
#
# ciecplib is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ciecplib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ciecplib.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`ciecplib.utils`."""

import os
from pathlib import Path
from unittest import mock

import pytest

from .. import utils as ciecplib_utils

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

EcpIdP = ciecplib_utils.EcpIdentityProvider

# mock response from ECP IdP list
INST_DICT = {
    "Institution A": "https://login.insta.org/idp/profile/SAML2/SOAP/ECP",
    "Institution A (Kerberos)":
        "https://login.insta.krb/idp-krb/profile/SAML2/SOAP/ECP",
    "Institution B": "https://login.instb.org/idp/profile/SAML2/SOAP/ECP",
    "Institution C": "https://login.instc.org/idp/profile/SAML2/SOAP/ECP",
    "Institution C (backup)":
        "https://login2.instc.org/idp/profile/SAML2/SOAP/ECP",
}
INSTITUTIONS = [
    EcpIdP(
        name,
        url,
        name.endswith("(Kerberos)"),
    ) for name, url in INST_DICT.items()
]
RAW_IDP_LIST = "\n".join(map("{0.url} {0.name}".format, INSTITUTIONS)).encode()


@mock.patch(
    "os.getlogin" if os.name == "nt" else "os.getuid",
    mock.MagicMock(return_value=123),
)
@mock.patch.dict("os.environ")
def test_get_ecpcookie_path():
    if os.name == "nt":
        path = r"C:\WINDOWS\Temp\ecpcookie.123"
    else:
        path = "/tmp/ecpcookie.u123"
    assert ciecplib_utils.get_ecpcookie_path() == Path(path)


@mock.patch(
    "os.getlogin" if os.name == "nt" else "os.getuid",
    mock.MagicMock(return_value=123),
)
def test_get_x509_proxy_path():
    os.environ.pop("X509_USER_PROXY", None)
    if os.name == "nt":
        path = r"C:\WINDOWS\Temp\x509up_123"
    else:
        path = "/tmp/x509up_u123"
    assert ciecplib_utils.get_x509_proxy_path() == Path(path)


def test_get_idps(requests_mock):
    requests_mock.get("https://idp-list-url", content=RAW_IDP_LIST)
    assert ciecplib_utils.get_idps("https://idp-list-url") == INSTITUTIONS


@pytest.mark.parametrize("value, krb, result", [
    # direct match for institution
    ("Institution A", False, INST_DICT["Institution A"]),
    ("Institution A (Kerberos)", None, INST_DICT["Institution A (Kerberos)"]),
    # kerberos selection
    ("Institution A", True, INST_DICT["Institution A (Kerberos)"]),
    # partial match for URL
    ("insta", False, INST_DICT["Institution A"]),
    ("login.insta.krb", None, INST_DICT["Institution A (Kerberos)"]),
    ("instb", None, INST_DICT["Institution B"]),
    # valid ECP URL, just return it
    ("https://myidp.org/idp/profile/SAML2/SOAP/ECP", None,
     "https://myidp.org/idp/profile/SAML2/SOAP/ECP"),
    # preferred match based on institution name
    ("Institution C", None, INST_DICT["Institution C"]),
    # preferred match based on URL
    ("instc", None, INST_DICT["Institution C"]),
])
def test_get_idp_url(requests_mock, value, krb, result):
    # only proxy idp-list-url if we need to
    if not value.endswith(r"/ECP"):
        requests_mock.get("https://idp-list-url", content=RAW_IDP_LIST)
    assert ciecplib_utils.get_idp_url(
        value,
        idplist_url="https://idp-list-url",
        kerberos=krb,
    ) == result


@pytest.mark.parametrize("inst", [
    # name too vague, multiple matches
    "Institution",
    # URL too vague, multiple matches
    "login",
    # no matches
    "something else",
])
def test_get_idp_url_error(requests_mock, inst):
    requests_mock.get("https://idp-list-url", content=RAW_IDP_LIST)
    with pytest.raises(ValueError):
        ciecplib_utils.get_idp_url(inst, idplist_url="https://idp-list-url")


@pytest.mark.parametrize("matches, out", (
    # match one primary
    ([EcpIdP("Inst (main)", "", False), EcpIdP("Inst (backup)", "", False)],
     [EcpIdP("Inst (main)", "", False)]),
    ([EcpIdP("Inst (primary)", "", False), EcpIdP("Inst (backup)", "", False)],
     [EcpIdP("Inst (primary)", "", False)]),
    ([EcpIdP("Inst 1", "", False), EcpIdP("Inst 2", "", False)],
     [EcpIdP("Inst 1", "", False)]),
    # match all-bar-one secondaries
    ([EcpIdP("Inst", "", False), EcpIdP("Inst 2", "", False)],
     [EcpIdP("Inst", "", False)]),
    # no match, return both
    ([EcpIdP("Inst A", "", False), EcpIdP("Inst B", "", False)],
     [EcpIdP("Inst A", "", False), EcpIdP("Inst B", "", False)]),
))
def test_preferred_match(matches, out):
    assert ciecplib_utils._preferred_match(matches) == out
